# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2017-2019 UBports
# Copyright (C) 2013-2015 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Calculator app autopilot tests."""

import os
import shutil
import logging
import fixtures

import ubuntu_calculator_app

from autopilot.testcase import AutopilotTestCase
from autopilot import logging as autopilot_logging

import ubuntuuitoolkit
from ubuntuuitoolkit import base

logger = logging.getLogger(__name__)


class CalculatorAppTestCase(AutopilotTestCase):
    """A common test case class that provides several useful methods for
    the ubuntu-calculator-app tests.

    """

    local_location = os.path.dirname(os.path.dirname(os.getcwd()))

    local_location_qml = os.path.join(local_location,
                                      'app/ubuntu-calculator-app.qml')

    installed_location_qml = os.path.join('/usr/share/ubuntu-calculator-app/',
                                          'ubuntu-calculator-app.qml')

    def get_launcher_and_type(self):
        if os.path.exists(self.local_location_qml):
            launcher = self.launch_test_local
            test_type = 'local'
        elif os.path.exists(self.installed_location_qml):
            launcher = self.launch_test_installed
            test_type = 'deb'
        else:
            launcher = self.launch_test_click
            test_type = 'click'
        return launcher, test_type

    def setUp(self):
        super(CalculatorAppTestCase, self).setUp()
        self.clear_calculator_database()
        self.launcher, self.test_type = self.get_launcher_and_type()
        if self.test_type is 'click':
            # just use home for now on devices
            temp_dir = os.environ.get('HOME')

            # before each test, remove the app's settings
            local_dir = os.path.join(temp_dir, '.config/com.ubuntu.calculator')

            if (os.path.exists(local_dir)):
                shutil.rmtree(local_dir)

        # Unset the current locale to ensure locale-specific data
        # (day and month names, first day of the week, …) doesn’t get
        # in the way of test expectations.
        self.useFixture(fixtures.EnvironmentVariable('LC_ALL', 'C'))
        self.app = ubuntu_calculator_app.CalculatorApp(self.launcher(),
                                                       self.test_type)

    @autopilot_logging.log_action(logger.info)
    def launch_test_local(self):
        return self.launch_test_application(
            base.get_qmlscene_launch_command(),
            self.local_location_qml,
            app_type='qt',
            emulator_base=ubuntuuitoolkit.UbuntuUIToolkitCustomProxyObjectBase)

    @autopilot_logging.log_action(logger.info)
    def launch_test_installed(self):
        return self.launch_test_application(
            base.get_qmlscene_launch_command(),
            self.installed_location_qml,
            app_type='qt',
            emulator_base=ubuntuuitoolkit.UbuntuUIToolkitCustomProxyObjectBase)

    @autopilot_logging.log_action(logger.info)
    def launch_test_click(self):
        return self.launch_click_package(
            "com.ubuntu.calculator",
            emulator_base=ubuntuuitoolkit.UbuntuUIToolkitCustomProxyObjectBase)

    def clear_calculator_database(self):
        calculator_database_path = os.path.join(
            os.path.expanduser('~'),
            '.local',
            'share',
            'com.ubuntu.calculator'
        )

        if os.path.exists(calculator_database_path):
            shutil.rmtree(calculator_database_path)
            lambda: os.path.exists(calculator_database_path).wait_for(False)



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2017-2019 UBports
# Copyright (C) 2013, 2016 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Calculator app autopilot emulators."""

from time import sleep
import ubuntuuitoolkit


def click_object(func):
    """Wrapper which clicks the returned object"""
    def func_wrapper(self, *args, **kwargs):
        return self.pointing_device.click_object(func(self, *args, **kwargs))

    return func_wrapper


class CalculatorApp(object):
    """Autopilot helper object for the calculator application."""

    def __init__(self, app_proxy, test_type):
        self.app = app_proxy
        self.test_type = test_type
        self.main_view = self.app.select_single(MainView)

    @property
    def pointing_device(self):
        return self.app.pointing_device


class CalculationHistory(object):
    """Helper object for the calculation history."""

    def __init__(self, app_proxy):
        self.app = app_proxy

    def contains(self, entry):
        found = False

        history = self.get_all_history_entries()
        for line in history:
            calculation = self._get_entry(line)
            if calculation == entry.strip():
                found = True
                break

        return found

    def get_all_history_entries(self):
        return self.app.select_many('QQuickRow', objectName='historyrow')

    def get_history_entry_count(self):
        return len(self.get_all_history_entries())

    def get_history_entry(self, entry):
        return self.app.select_single('QQuickText',
                                      objectName='result' + str(entry))

    def get_formula_entry(self, entry=0):
        return self.app.select_single('QQuickText',
                                      objectName='formula' + str(entry))

    def get_formula_value(self, entry=0):
        return self.get_formula_entry(entry).text.strip()

    def get_result_value(self, entry=0):
        return self.get_history_entry(entry).text.strip()

    def _get_entry(self, line):
        values = line.select_many('QQuickText')
        result = ''
        formula = ''
        for value in values:
            if 'result' in value.objectName:
                result = value.text.strip()
            elif 'formula' in value.objectName:
                formula = value.text.replace(" ", "")

        return formula + result


class MainView(ubuntuuitoolkit.MainView):
    """Calculator MainView Autopilot emulator."""

    BUTTONS = {'delete': 'deleteButton', '*': 'multiplyButton',
               '/': 'divideButton', '.': 'pointButton',
               '=': 'equalsButton', '-': 'minusButton', '+': 'plusButton',
               '0': 'zeroButton', '1': 'oneButton', '2': 'twoButton',
               '3': 'threeButton', '4': 'fourButton', '5': 'fiveButton',
               '6': 'sixButton', '7': 'sevenButton', '8': 'eightButton',
               '9': 'nineButton', 'bracket': 'universalBracketButton',
               'square': 'squareButton', 'cube': 'cubeButton',
               'power': 'powerButton', 'log': 'logarithmButton',
               'e': 'eNumberButton', '!': 'factorialNumberButton',
               'i': 'iButton', 'sin': 'sinusButton', 'cos': 'cosButton'}

    def __init__(self, *args):
        super(MainView, self).__init__(*args)
        self.visible.wait_for(True)

    def insert(self, expression):
        for operand in expression:
            self.press(operand)

    def press_universal_bracket(self):
        self.press('bracket')

    def delete(self):
        self.press('delete')

    def clear(self):
        self.press_and_hold('delete')

    def press_and_hold(self, button):
        button = self.wait_select_single('KeyboardButton',
                                         objectName=MainView.BUTTONS[button])

        self.pointing_device.move_to_object(button)
        self.pointing_device.press()
        button.pressed.wait_for(True)
        sleep(3)
        self.pointing_device.release()

    def press(self, button):
        button = self.wait_select_single('KeyboardButton',
                                         objectName=MainView.BUTTONS[button])

        self.pointing_device.move_to_object(button)
        # we use press and release so we can check the qml property
        # and ensure the button is pressed long enough to be recieved
        # and processed correctly
        # using a larger press_duration for click_object would be inferior
        # as it would cause longer delays (we are forced to arbitrarily decide
        # how long to press each time) and potentially fail.
        # balloons 2015-01-29
        self.pointing_device.press()
        # this sleeps represents our minimum press time,
        # should button.pressed be true without any wait
        sleep(0.1)
        button.pressed.wait_for(True)
        self.pointing_device.release()

    def get_history(self):
        history = self.wait_select_single('ScrollableView',
                                          objectName='scrollableView')

        return CalculationHistory(history)

    def get_result(self):
        return self.wait_select_single('TextField',
                                       objectName='textInputField').displayText

    def get_walkthrough_page(self):
        return self.wait_select_single('Walkthrough',
                                       objectName="walkthroughPage")

    def show_scientific_keyboard(self):
        self._scientific_keyboard()

    def hide_scientific_keyboard(self):
        self._scientific_keyboard(enable=False)

    def _scientific_keyboard(self, enable=True):
        y = (self.globalRect[1] + self.globalRect[3] / 2) + 150

        x_middle = self.globalRect[0] + self.globalRect[2] / 2

        if enable:
            x_start = x_middle + 150
            x_stop = x_middle - 150
        else:
            x_start = x_middle - 150
            x_stop = x_middle + 150

        self.pointing_device.drag(x_start, y, x_stop, y)

        # TODO: Find a better implementation to avoid this, if possible.
        sleep(2)

    def enter_text_via_keyboard(self, textToEnter):
        """Enter specific formula with keyboard

        :param text: text to enter via keyboard

        """
        ubuntuuitoolkit.get_keyboard().type(textToEnter, delay=0.1)

    def press_and_release_key(self, keyToPress):
        """Press and release specific key

        :param keyToPress: key to press

        """
        ubuntuuitoolkit.get_keyboard().press_and_release(keyToPress, delay=0.1)


class Page(ubuntuuitoolkit.UbuntuUIToolkitCustomProxyObjectBase):
    """Autopilot helper for Pages."""
    def __init__(self, *args):
        super(Page, self).__init__(*args)


class Walkthrough(Page):
    """ Autopilot helper for the walkthrough page """
    def __init__(self, *args):
        super(Walkthrough, self).__init__(*args)

        self.visible.wait_for(True)

    @click_object
    def skip(self):
        return self.wait_select_single("UCLabel", objectName="skipLabel")



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

"""Calculator app autopilot tests."""

from autopilot.matchers import Eventually
from testtools.matchers import Equals

from ubuntu_calculator_app.tests import CalculatorAppTestCase
from autopilot.platform import model
import unittest

class MainTestCase(CalculatorAppTestCase):

    def setUp(self):
        super(MainTestCase, self).setUp()
        self.app.main_view.get_walkthrough_page().skip()

    @unittest.skipIf(model() is not 'Desktop',
                     "Keyboard test is supported only on Desktop")
    def test_simple_calculation_via_keyboard(self):
        self.app.main_view.enter_text_via_keyboard('.9')
        self._assert_result_is(u'0.9')

        self.app.main_view.enter_text_via_keyboard('*9/')
        self._assert_result_is(u'0.9×9÷')

        self.app.main_view.enter_text_via_keyboard('.3=')
        self._assert_result_is(u'27')
        self._assert_history_contains(u'0.9×9÷0.3=27')

        self.app.main_view.enter_text_via_keyboard('+3')
        self.app.main_view.press_and_release_key('Enter')
        self._assert_result_is(u'30')
        self._assert_history_contains(u'0.9×9÷0.3=27')

    def test_add_operator_after_result(self):
        self.app.main_view.insert('9*9=')

        self._assert_result_is(u'81')
        self._assert_history_contains(u'9×9=81')
        # We expect that after inserting operator,
        # the result will be used in next calculation
        self.app.main_view.insert('+9=')

        self._assert_result_is(u'90')
        self._assert_history_contains(u'81+9=90')

    def test_enter_number_after_result(self):
        self.app.main_view.insert('3*3=')

        self._assert_result_is(u'9')

        # We expect that after inserting number,
        # result will be deleted
        self._assert_history_contains(u'3×3=9')
        self.app.main_view.insert('2+3=')

        self._assert_result_is(u'5')
        self._assert_history_contains(u'2+3=5')

    def test_temporarly_result(self):

        self.app.main_view.press_universal_bracket()
        self.app.main_view.insert('2450.1*369')
        self.app.main_view.press_universal_bracket()
        self.app.main_view.insert('+')
        self._assert_result_is(u'904086.9+')
        self.app.main_view.insert('3.1=')

        self._assert_result_is(u'904090')
        self._assert_history_contains(u'(2450.1×369)+3.1=904090')

    def test_addding_operator_after_calculation(self):
        self.app.main_view.insert('8*8.1=')

        self._assert_result_is(u'64.8')
        self._assert_history_contains(u'8×8.1=64.8')

        self.app.main_view.insert('+5.2=')

        self._assert_result_is(u'70')
        self._assert_history_contains(u'64.8+5.2=70')

    def test_addding_number_after_calculation(self):
        self.app.main_view.insert('3*3.1=')

        self._assert_result_is(u'9.3')
        self._assert_history_contains(u'3×3.1=9.3')

        self.app.main_view.insert('8-7=')

        self._assert_result_is(u'1')
        self._assert_history_contains(u'8−7=1')

    def test_operation_after_clear(self):
        self.app.main_view.insert('7*7')

        self._assert_result_is(u'7×7')

        self.app.main_view.clear()
        self._assert_result_is(u'')
        self.app.main_view.insert('2*0=')

        self._assert_result_is(u'0')
        self._assert_history_contains(u'2×0=0')

    def test_operation_after_delete(self):
        self.app.main_view.insert('8*8=')

        self._assert_result_is(u'64')
        self._assert_history_contains(u'8×8=64')

        self.app.main_view.delete()
        self._assert_result_is(u'')
        self.app.main_view.insert('9*9=')

        self._assert_result_is(u'81')
        self._assert_history_contains(u'9×9=81')

    def test_small_numbers(self):
        self.app.main_view.insert('0.0000000001+1=')
        self._assert_result_is(u'1.0000000001')
        self._assert_history_contains(u'0.0000000001+1=1.0000000001')

        self.app.main_view.delete()

        self.app.main_view.insert('0.0000000001/10=')
        self._assert_result_is(u'1e−11')
        self._assert_history_contains(u'0.0000000001÷10=1e−11')

    def test_operation_on_large_numbers(self):
        self.app.main_view.insert('99999999999*99999999999=')
        self._assert_result_is(u'9.9999999998e+21')
        self._assert_history_contains(u'99999999999×99999999999='
                                      '9.9999999998e+21')

        self.app.main_view.insert('*100=')

        self._assert_result_is(u'9.9999999998e+23')
        self._assert_history_contains(u'9.9999999998e+21×100='
                                      '9.9999999998e+23')

    def test_brackets_precedence(self):
        self.app.main_view.insert('2*')
        self.app.main_view.press_universal_bracket()
        self.app.main_view.insert('3+4')
        self.app.main_view.press_universal_bracket()
        self.app.main_view.insert('=')
        self._assert_result_is(u'14')
        self._assert_history_contains(u'2×(3+4)=14')

        self.app.main_view.delete()
        self.app.main_view.insert('4')
        self.app.main_view.press_universal_bracket()
        self.app.main_view.insert('3-2')
        self.app.main_view.press_universal_bracket()
        self._assert_result_is(u'4×(3−2)')
        self.app.main_view.insert('=')
        self._assert_result_is(u'4')
        self._assert_history_contains(u'4×(3−2)=4')

    def test_operators_precedence(self):
        self.app.main_view.insert('2+2*2=')

        self._assert_result_is(u'6')
        self._assert_history_contains(u'2+2×2=6')

        self.app.main_view.delete()
        self.app.main_view.insert('2-2*2=')

        self._assert_result_is(u'−2')
        self._assert_history_contains(u'2−2×2=−2')

        self.app.main_view.delete()
        self.app.main_view.insert('5+6/2=')

        self._assert_result_is(u'8')
        self._assert_history_contains(u'5+6÷2=8')

    def test_divide_with_zero(self):
        self.app.main_view.insert('0/5=')

        self._assert_result_is(u'0')
        self._assert_history_contains(u'0÷5=0')

    def test_divide_by_zero(self):
        self.app.main_view.insert('5/0=')

        self._assert_result_is(u'\u221e')
        self._assert_history_contains(u'5÷0=\u221e')

    def test_divide_zero_by_zero(self):
        self.app.main_view.insert('0/0=')

        self._assert_result_is(u'NaN')
        self._assert_history_contains(u'0÷0=NaN')

    def test_equals_doesnt_change_numbers(self):
        count_one = self.app.main_view.get_history().get_history_entry_count()

        self.app.main_view.insert('5*5=')
        self._assert_result_is(u'25')
        count_two = self.app.main_view.get_history().get_history_entry_count()
        self.assertTrue(count_one == count_two - 1)

        self.app.main_view.insert('===')
        self._assert_result_is(u'25')
        count_three = self.app.main_view.get_history() \
                          .get_history_entry_count()
        self.assertTrue(count_two == count_three)

    def test_divide_with_infinite_number_as_result(self):
        self.app.main_view.insert('1/3=')
        self._assert_result_is(u'0.333333333333')
        self._assert_history_contains(u'1÷3=0.333333333333')

    def test_operation_on_infinite_number(self):
        self.app.main_view.insert('5/3=')
        self._assert_result_is(u'1.666666666667')
        self._assert_history_contains(u'5÷3=1.666666666667')

        self.app.main_view.insert('-1=')
        self._assert_result_is(u'0.666666666667')

    def test_adding_comma_without_number(self):
        # Validation of the decimal separator
        # We are trying to add several commas into one number
        # Only first comma in the number should be allowed
        self.app.main_view.insert('..1.3*.*5.=')
        self._assert_result_is(u'0.065')
        self._assert_history_contains(u'0.13×0.5=0.065')
        self.app.main_view.insert('.7')
        self._assert_result_is(u'0.7')

    def test_adding_comma_without_number_on_temp_result(self):

        self.app.main_view.press_universal_bracket()
        self.app.main_view.insert('3+6*9')
        self._assert_result_is(u'(3+6×9')
        self.app.main_view.press_universal_bracket()
        self.app.main_view.insert('+')
        self._assert_result_is(u'57+')
        self.app.main_view.insert('.1')
        self._assert_result_is(u'57+0.1')
        self.app.main_view.insert('=')
        self._assert_history_contains(u'(3+6×9)+0.1=57.1')
        self._assert_result_is(u'57.1')

    def test_square(self):
        self.app.main_view.insert('4')
        self.app.main_view.show_scientific_keyboard()
        self.app.main_view.press('square')
        self.app.main_view.hide_scientific_keyboard()
        self.app.main_view.insert('=')

        self._assert_result_is(u'16')
        self._assert_history_contains(u'4^2=16')

    def test_adding_square_to_empty_formula(self):
        self.app.main_view.show_scientific_keyboard()
        self.app.main_view.press('square')
        self._assert_result_is(u'')

    def test_adding_square_after_operator(self):
        self.app.main_view.insert('6/')
        self.app.main_view.show_scientific_keyboard()
        self.app.main_view.press('square')
        self._assert_result_is(u'6÷')

    def test_cube(self):
        self.app.main_view.insert('3')
        self.app.main_view.show_scientific_keyboard()
        self.app.main_view.press('cube')
        self.app.main_view.hide_scientific_keyboard()
        self.app.main_view.insert('=')

        self._assert_result_is(u'27')
        self._assert_history_contains(u'3^3=27')

    def test_power(self):
        self.app.main_view.insert('2')
        self.app.main_view.show_scientific_keyboard()
        self.app.main_view.press('power')
        self.app.main_view.hide_scientific_keyboard()
        self.app.main_view.insert('-4=')

        self._assert_result_is(u'0.0625')
        self._assert_history_contains(u'2^−4=0.0625')

    def test_loge(self):
        self.app.main_view.show_scientific_keyboard()
        self.app.main_view.press('log')
        self.app.main_view.press('e')
        self.app.main_view.hide_scientific_keyboard()
        self.app.main_view.insert('=')

        self._assert_result_is(u'1')

    def test_factorial(self):
        self.app.main_view.insert('4')
        self.app.main_view.show_scientific_keyboard()
        self.app.main_view.press('!')
        self.app.main_view.hide_scientific_keyboard()
        self.app.main_view.insert('=')

        self._assert_result_is(u'24')
        self._assert_history_contains(u'4!=24')

    def test_factorial_with_brackets(self):
        self.app.main_view.press_universal_bracket()
        self.app.main_view.insert('3')
        self.app.main_view.show_scientific_keyboard()
        self.app.main_view.press('!')
        self.app.main_view.hide_scientific_keyboard()
        self.app.main_view.insert('*2')
        self.app.main_view.press_universal_bracket()
        self.app.main_view.insert('=')

        self._assert_result_is(u'12')
        self._assert_history_contains(u'(3!×2)=12')

    def test_sin(self):
        self.app.main_view.show_scientific_keyboard()
        self.app.main_view.press('sin')
        self.app.main_view.hide_scientific_keyboard()
        self.app.main_view.insert('0=')

        self._assert_result_is(u'0')
        self._assert_history_contains(u'sin(0)=0')

    def test_cos(self):
        self.app.main_view.show_scientific_keyboard()
        self.app.main_view.press('cos')
        self.app.main_view.hide_scientific_keyboard()
        self.app.main_view.insert('0')
        self.app.main_view.press_universal_bracket()
        self.app.main_view.insert('=')

        self._assert_result_is(u'1')
        self._assert_history_contains(u'cos(0)=1')

    def test_validation_complex_numbers(self):
        self.app.main_view.insert('66')
        self.app.main_view.show_scientific_keyboard()
        self.app.main_view.press('i')
        self.app.main_view.hide_scientific_keyboard()
        self.app.main_view.insert('*')
        self.app.main_view.show_scientific_keyboard()
        self.app.main_view.press('i')
        self.app.main_view.press('i')
        self.app.main_view.press('i')
        self._assert_result_is(u'66i×i')
        self.app.main_view.hide_scientific_keyboard()
        self.app.main_view.insert('33=')
        self._assert_result_is(u'−66')
        self._assert_history_contains(u'66i×i=−66')

    def test_formatting_long_complex_numbers(self):
        self.app.main_view.press_universal_bracket()
        self.app.main_view.insert('3+4')
        self.app.main_view.show_scientific_keyboard()
        self.app.main_view.press('i')
        self.app.main_view.hide_scientific_keyboard()
        self.app.main_view.press_universal_bracket()
        self._assert_result_is(u'(3+4i)')
        self.app.main_view.show_scientific_keyboard()
        self.app.main_view.press('square')
        self.app.main_view.hide_scientific_keyboard()
        self.app.main_view.insert('=')
        self._assert_result_is(u'−6.999999999999997+24i')
        self._assert_history_contains(u'(3+4i)^2=−6.999999999999997+24i')

    def test_floating_point_round_error(self):
        self.app.main_view.insert('0.1+0.2=')
        self._assert_result_is(u'0.3')
        self._assert_history_contains(u'0.1+0.2=0.3')

    def _assert_result_is(self, value):
        self.assertThat(self.app.main_view.get_result,
                        Eventually(Equals(value)))

    def _assert_history_contains(self, value):
        self.assertTrue(self.app.main_view.get_history().contains(value))
